import os
import fcntl
import types
import sys
import tempfile
import uuid
import code
import traceback
import threading
import signal
import signalfd
import operator
import re
import builtins        
import functools

from .. import iterio
from .. import redir

class PipelineError(Exception):
    description = "Pipeline"
    def __init__(self, pipeline):
        self.pipeline = pipeline
    def __str__(self):
        return "%s: %s:\n\n%s" % (
            self.description,
            self.pipeline,
            "\n\n================================\n\n".join(
                [proc.__repr__(display_output=True)
                 for proc in self.pipeline.failed_processes]))

class PipelineFailed(PipelineError, Exception): description = "Pipeline failed"
class PipelineInterrupted(PipelineError, KeyboardInterrupt): description = "Pipeline canceled"

class RunningPipeline(object):
    def __init__(self, processes, pipeline):
        self.processes = processes
        self.pipeline = pipeline
    def __iter__(self):
        def handle_input():
            for line in iterio.LineInputHandler(self.pipeline._redirects.stdout.pipe, usage=self):
                yield line
            self.wait()
        return iter(handle_input())
    def wait(self):
        while functools.reduce(operator.__or__, (proc.is_running for proc in self.processes), False):
            iterio.get_io_manager().handle_io()
        if self.failed_processes:
            raise PipelineFailed(self)
    @property
    def failed_processes(self):
        return [proc
                for proc in self.processes
                if not proc.is_running and proc.is_failed]
    def remove_output_files(self):
        for proc in self.pipeline.processes:
            proc.remove_output_files()
    def __repr__(self):
        return repr(self.pipeline)

class RunningItem(object):
    def __init__(self, cmd, iohandler):
        self.cmd = cmd
        self.iohandler = iohandler
        self.output_content = {}
    @property
    def is_running(self):
        return self.iohandler.is_running
    def handle_finish(self):
        for fd, redirect in self.cmd._redirects.redirects.items():
            if not isinstance(redirect.pipe, redir.STRING): continue
            with open(redirect.pipe.path) as f:
                self.output_content[fd] = f.read()
            os.unlink(redirect.pipe.path)
    @property
    def output_files(self):
        if self.output_content is not None: return {}
        return {fd: redirect.pipe
                for fd, redirect in self.cmd._redirects.redirects.items()
                if isinstance(redirect.pipe, redir.TMP) and not isinstance(redirect.pipe, redir.STRING)}
    def remove_output_files(self):
        for fd, name in self.output_files.items():
            os.unlink(name)
    def __getattr__(self, name):
        return getattr(self.iohandler, name)

class RunningFunction(RunningItem):
    @property
    def is_failed(self):
        return self.iohandler.exception is not None
    def __repr__(self, display_output=False):
        status = list(self.iohandler._repr_args())
        if self.iohandler.exception is not None:
            status.append(str(self.iohandler.exception))
        if status:
            status = ' (' + ', '.join(status) + ')'
        else:
            status = ''
        return '%s%s' % (self.cmd._function_name(), status)

class RunningProcess(RunningItem):
    class ProcessSignalHandler(iterio.ProcessSignalHandler):
        def __init__(self, process, pid):
            self.process = process
            iterio.ProcessSignalHandler.__init__(self, pid)
        def handle_event(self, event):
            res = iterio.ProcessSignalHandler.handle_event(self, event)
            if not self.is_running:
                self.process.handle_finish()
            return res
    def __init__(self, cmd, pid):
        RunningItem.__init__(self, cmd, self.ProcessSignalHandler(self, pid))
    @property
    def is_failed(self):
        return self.iohandler.last_event["ssi_status"] != 0
    def __repr__(self, display_output=False):
        status = []
        last_event = self.iohandler.last_event
        if last_event:
            last_event = iterio.siginfo_to_names(last_event)
        if self.iohandler.is_running:
            if last_event and last_event["ssi_code"] != 'CLD_CONTINUED':
                status.append("status=%s" % self.iohandler.last_event["ssi_code"])
                status.append("signal=%s" % self.iohandler.last_event["ssi_status"])
        else:
            status.append("exit_code=%s" % self.iohandler.last_event["ssi_status"])
            if len(self.output_files):
                for fd, output_file in self.output_files.items():
                    status.append("%s=%s" % (fd, output_file))
        if status:
            status = ' (' + ', '.join(status) + ')'
        else:
            status = ''
        res = '%s%s' % (self.iohandler.pid, status)
        if display_output:
            res += "\n"
            for fd, value in self.output_content.items():
                fd = redir.Redirect.names_to_fd.get(fd, fd)
                res += "%s content:\n%s\n" % (fd, value) 
        return res



# Minimal reimplementation of copy.deepcopy that happily copies
# objects inheriting from type...
import threading
import contextlib

copy_state = threading.local()

@contextlib.contextmanager
def copy_session(sess = None):
    old_memo = getattr(copy_state, "memo", None)
    if sess:
        copy_state.memo = sess
    else:
        copy_state.memo = old_memo or {}
    try:
        yield copy_state.memo
    finally:
        copy_state.memo = old_memo

def deepcopy(obj):
    with copy_session() as memo:
        key = id(obj)
        if key not in memo:
            if isinstance(obj, (list, tuple)):
                memo[key] = type(obj)(deepcopy(item) for item in obj)
            elif isinstance(obj, dict):
                memo[key] = type(obj)({deepcopy(key):deepcopy(value) for key, value in obj.items()})
            elif isinstance(obj, bytes):
                memo[key] = bytes(obj)
            elif isinstance(obj, str):
                memo[key] = str(obj)
            elif hasattr(obj, "__deepcopy__"):
                memo[key] = obj.__deepcopy__()
            else:
                memo[key] = obj
        return memo[key]



import os
import fcntl
import types
import sys
import tempfile
import uuid
import code
import threading

from . import pipe
from . import log
from . import copy

try:
    MAXFD = os.sysconf("SC_OPEN_MAX")
except:
    MAXFD = 256

class SpecialRedirect(object):
    class __metaclass__(type):
        def __str__(self):
            return repr(self)
        def __repr__(self):
            return self.__name__
    def __init__(self, **kws):
        object.__setattr__(self, "kws", kws)
    def __getattr__(self, name):
        return self.kws[name]
    def __setattr__(self, name):
        self.kws[name] = value
    def __str__(self):
        return repr(self)
    def __repr__(self):
        return "%s(%s)" % (
            type(self),
            ",".join(("%s=%s" % (name, value)
                      for name, value in self.kws.items())))

class PIPE(SpecialRedirect): pass
class TMP(SpecialRedirect): pass
class STRING(TMP): pass

def flags_to_string(flags):
    return ",".join([name[2:]
                     for name in dir(os)
                     if name.startswith("O_") and flags & getattr(os, name)])

class Redirect(object):
    fd_names = {"stdin": 0, "stdout": 1, "stderr": 2}
    names_to_fd = {value: key for key, value in fd_names.items()}
    fd_flags = {
        0: os.O_RDONLY,
        1: os.O_WRONLY | os.O_CREAT,
        2: os.O_WRONLY | os.O_CREAT
        }
    def __init__(self, fd, source = None, flag = None, mode = 0o777, pipe=None, borrowed=False):
        if isinstance(fd, Redirect):
            fd, source, flag, mode, pipe, borrowed = fd.fd, fd.source, fd.flag, fd.mode, fd.pipe, fd.borrowed
        if not isinstance(fd, int):
            fd = self.fd_names[fd]
        if flag is None:
            flag = self.fd_flags[fd]
        self.fd = fd
        self.source = source
        self.flag = flag
        self.mode = mode
        self.pipe = pipe
        self.borrowed = borrowed
    def borrow(self):
        self.borrowed = True
    def __deepcopy__(self, memo = {}):
        return type(self)(self.fd, copy.deepcopy(self.source), self.flag, self.mode, self.pipe, self.borrowed)

    def open(self, borrow = True):
        source = self.source
        if not isinstance(source, int):
            log.log("Opening %s in %s for %s" % (self.source, self.flag, self.fd), "fd")
            source = os.open(source, self.flag, self.mode)
            log.log("Done opening %s in %s for %s" % (self.source, self.flag, self.fd), "fd")
        elif not borrow:
            log.log("Opening %s for %s" % (self.source, self.fd), "fd")
            source = os.dup(source)
        return source
    def close_source_fd(self):
        # FIXME: Only close source fds that come from pipes instead of this hack...
        if isinstance(self.source, int) and not self.borrowed:
            log.log("CLOSE SOURCE %s" % (self.source,), "fd")
            os.close(self.source)
    def perform(self):
        source = self.open()
        assert source != self.fd
        log.log("perform dup2(%s, %s)" % (source, self.fd), "fd")
        os.dup2(source, self.fd)
        log.log("perform close(%s)" % (source), "fd")
        os.close(source)
    def move(self, fd):
        self = Redirect(self)
        if isinstance(self.source, int):
            log.log("move dup2(%s, %s)" % (self.source, fd), "fd")
            os.dup2(self.source, fd)
            log.log("move close(%s)" % (self.source), "fd")
            os.close(self.source)
            self.source = fd
        return self
    def make_pipe(self):
        if isinstance(self.source, type) and issubclass(self.source, PIPE):
            rfd, wfd = pipe.pipe_cloexec()
            if self.flag & os.O_WRONLY:
                sourcefd, pipefd = wfd, rfd
            else:
                pipefd, sourcefd = wfd, rfd
            return type(self)(self.fd, sourcefd, self.flag, self.mode, pipefd)
        elif isinstance(self.source, type) and issubclass(self.source, (TMP, STRING)):
            if not (self.flag & os.O_WRONLY):
                raise Exception("Invalid flag for %s redirect - must be O_WRONLY" % self.source)
            sourcefd, pipefd = tempfile.mkstemp()
            return type(self)(self.fd, sourcefd, self.flag, self.mode, self.source(path=pipefd))
        else:
            return self

    def __repr__(self):
        flagmode = []
        if self.flag != self.fd_flags.get(self.fd, None):
            flagmode.append(flags_to_string(self.flag))
        if self.mode != 0o777:
            flagmode.append("m=%s" % self.mode)
        if flagmode:
            flagmode = "[" + ",".join(flagmode) + "]"
        else:
            flagmode = ""
        if self.flag & os.O_WRONLY:
            arrow = "-%s->" % flagmode
        else:
            arrow = "<-%s-" % flagmode
        arrow = "%s %s %s" % (self.fd, arrow, self.source)
        if self.borrowed:
            arrow = "(%s)" % arrow
        items = [arrow]
        if self.pipe is not None:
            items.append("pipe=%s" % self.pipe)
        return " ".join(items)

class Redirects(object):
    def __init__(self, *redirects, **kw):
        self.redirects = {}
        for redirect in redirects:
            if isinstance(redirect, Redirects):
                for item in redirect.redirects.values():
                    self.register(Redirect(item))
            else:
                self.register(Redirect(redirect))
    def borrow(self):
        for redirect in self.redirects.values():
            redirect.borrow()
    def register(self, redirect):
        if isinstance(redirect, Redirects):
            for item in redirect.redirects.values():
                self.register(item)
        else:
            if not isinstance(redirect, Redirect):
                redirect = Redirect(redirect)
            if redirect.source is None:
                del self.redirects[redirect.fd]
            else:
                self.redirects[redirect.fd] = redirect
        return self
    def redirect(self, *arg, **kw):
        self.register(Redirect(*arg, **kw))
        return self
    def merge(self, other):
        return Redirects(self, other)
    def find_free_fd(self):
        return max([redirect.fd
                    for redirect in self.redirects.values()]
                   + [redirect.source
                      for redirect in self.redirects.values()
                      if isinstance(redirect.source, int)]
                   + [2]) + 1
    def make_pipes(self):
        return type(self)(*[redirect.make_pipe()
                            for redirect in self.redirects.values()])
    def move_existing_fds(self):
        new_fd = self.find_free_fd()
        redirects = []
        for redirect in self.redirects.values():
            redirects.append(redirect.move(new_fd))
            new_fd += 1
        log.log("After move: %s" % repr(Redirects(*redirects)), "fd")
        return redirects
    def perform(self):
        try:
            for redirect in self.move_existing_fds():
                redirect.perform()
            self.close_other_fds()
        except Exception as e:
            import traceback
            log.log(e, "fd")
            log.log(traceback.format_exc(), "fd")
    def close_other_fds(self):
        # FIXME: Use os.closerange if available
        for i in xrange(0, MAXFD):
            if i in self.redirects: continue
            if i == log.logfd: continue
            try:
                os.close(i)
            except:
                pass
            else:
                log.log("CLOSE OTHER FDS %s" % (i,), "fd")
    def close_source_fds(self):
        for redirect in self.redirects.values():
            redirect.close_source_fd()
    def __getattr__(self, name):
        try:
            return self.redirects[Redirect.fd_names[name]]
        except KeyError:
            raise AttributeError(name)
    @classmethod
    def _coerce(cls, thing, direction):
        if thing is None:
            thing = "/dev/null"
        if isinstance(thing, (str, unicode)):
            thing = Redirect(direction, thing)
        if isinstance(thing, Redirect):
            thing = Redirects(thing)
        if not isinstance(thing, Redirects):
            raise ValueError(type(thing))
        return thing
    def __repr__(self):
        redirects = sorted(self.redirects.values(), key = lambda a: a.fd)
        return ", ".join(repr(redirect) for redirect in redirects)



import os
import fcntl
import select

def _set_cloexec_flag(fd, cloexec=True):
    try:
        cloexec_flag = fcntl.FD_CLOEXEC
    except AttributeError:
        cloexec_flag = 1

    old = fcntl.fcntl(fd, fcntl.F_GETFD)
    if cloexec:
        fcntl.fcntl(fd, fcntl.F_SETFD, old | cloexec_flag)
    else:
        fcntl.fcntl(fd, fcntl.F_SETFD, old & ~cloexec_flag)

def pipe_cloexec():
    """Create a pipe with FDs set CLOEXEC."""
    # Pipes' FDs are set CLOEXEC by default because we don't want them
    # to be inherited by other subprocesses: the CLOEXEC flag is removed
    # from the child's FDs by _dup2(), between fork() and exec().
    # This is not atomic: we would need the pipe2() syscall for that.
    r, w = os.pipe()
    _set_cloexec_flag(r)
    _set_cloexec_flag(w)
    return r, w



import json

def map(func):
    def apply_map(iter):
        for item in iter:
            if item is None:
                yield None
            else:
                yield func(item)
    apply_map.__name__ = "map(%s)" % (func.__name__,)
    return apply_map

def filter(func):
    def apply_filter(iter):
        for item in iter:
            if item is None:
                yield None
            elif func(item):
                yield item
    apply_map.__name__ = "filter(%s)" % (func.__name__,)
    return apply_filter

from_json = map(json.loads)
to_json = map(json.dumps)



import jedi.api

def jedi_init(self, source, namespaces, **kwds):
    super(jedi.api.Interpreter, self).__init__(source, **kwds)
    self.namespaces = namespaces

    self._parser = jedi.api.UserContextParser(self._grammar, self.source,
                                              self._orig_path, self._pos,
                                              self._user_context, self._parsed_callback,
                                              use_fast_parser=False)
    jedi.api.interpreter.add_namespaces_to_parser(self._evaluator, namespaces,
                                                  self._parser.module())

jedi.api.Interpreter.__init__ = jedi_init



import os
import fcntl
import select
import threading
import signalfd
import signal
import errno
from . import log

class RecursiveEvent(Exception): pass

def events_to_str(events):
    return ",".join([name for name in dir(select)
                     if (    name.startswith("POLL")
                         and events & getattr(select, name) != 0)])


class IOManager(object):
    def __init__(self):
        self.io_handlers = {}
        self.delay = 0

        self.poll = select.poll()

        self.cleanup = []

    def register_cleanup(self, cleanup):
        self.cleanup.append(cleanup)

    def delay_cleanup(self):
        self.delay += 1

    def perform_cleanup(self):
        if self.delay > 0:
            self.delay -= 1
        self._do_cleanup()

    def register(self, io_handler):
        self.enable(io_handler)
        self.io_handlers[io_handler.fd] = io_handler
        log.log("REGISTER %s, %s, %s" % (io_handler.fd, events_to_str(io_handler.events), io_handler), "ioreg")

    def enable(self, io_handler):
        self.poll.register(io_handler.fd, io_handler.events)

    def disable(self, io_handler):
        self.poll.unregister(io_handler.fd)

    def _do_cleanup(self):
        while self.delay == 0 and not len(self.io_handlers) and self.cleanup:
            self.cleanup.pop()()

    def deregister(self, io_handler):
        self.disable(io_handler)
        del self.io_handlers[io_handler.fd]
        self._do_cleanup()
        log.log("DEREGISTER %s, %s" % (io_handler.fd, io_handler), "ioreg")
    
    def handle_io(self, timeout = None):
        while self.io_handlers:
            try:
                events = self.poll.poll(timeout)
            except IOError as e:
                if e.errno == errno.EINTR:
                    continue
                raise

            log.log("EVENTS %s" % (", ".join("%s:%s" % (fd, events_to_str(event))
                                             for (fd, event) in events),), "ioevent")
            assert timeout is not None or events
            done = False
            recursive = False
            for fd, event in events:
                # Check if the fd is still registered. If we have
                # multiple events for the same fd, this might not be
                # the case for the second event...
                if fd in self.io_handlers:
                    try:
                        event_done = self.io_handlers[fd].handle_event(event)
                        done = done or event_done
                    except RecursiveEvent:
                        recursive = True
            if done:
                return
            if recursive:
                raise RecursiveEvent
            if timeout is not None:
                return

    def __repr__(self):
        return "IOManager(%s)" % (", ".join(repr(iohandler) for iohandler in self.io_handlers.values()),)

io_managers = threading.local()
def get_io_manager():
    global signal_manager
    if not hasattr(io_managers, 'manager'):
        io_managers.manager = IOManager()
        if isinstance(threading.current_thread(), threading._MainThread):
            signal_manager = SignalManager()
    return io_managers.manager

signal_manager = None
def get_signal_manager():
    global signal_manager
    get_io_manager()
    return signal_manager


class IOHandler(object):
    events = 0
    def __init__(self, fd, borrowed = False, usage = None):
        self.fd = fd
        self.borrowed = borrowed
        self.enabled = True
        self.usage = usage
        get_io_manager().register(self)
    def handle_event(self, event):
        pass
    def destroy(self):
        get_io_manager().deregister(self)
        if self.borrowed:
            log.log("HAND BACK %s, %s" % (self.fd, self), "ioreg")
        else:
            log.log("CLOSE DESTROY %s, %s" % (self.fd, self), "ioreg")
            os.close(self.fd)
    def enable(self):
        self.enabled = True
        get_io_manager().enable(self)
    def disable(self):
        self.enabled = False
        get_io_manager().enable(self)
    def _repr_args(self):
        args = [str(self.fd)]
        if self.usage:
            args.append("for %s" % repr(self.usage))
        if self.enabled:
            args.append("enabled")
        return args
    def __repr__(self):
        t = type(self)
        return "%s.%s(%s)" % (t.__module__, t.__name__, ",".join(self._repr_args()))

class OutputHandler(IOHandler):
    events = select.POLLOUT

    def __init__(self, fd, iter, borrowed = False, usage = None):
        self.iter = iter
        self.is_running = True
        self.recursion = False
        self.exception = None
        IOHandler.__init__(self, fd, borrowed, usage)

    def destroy(self):
        self.is_running = False
        IOHandler.destroy(self)

    def handle_event(self, event):
        if self.recursion:
            raise RecursiveEvent
        self.recursion = True
        try:
            return self.handle_event_non_recursive(event)
        finally:
            self.recursion = False

    def handle_event_non_recursive(self, event):
        try:
            val = next(self.iter)
            if val is not None:
                os.write(self.fd, val)
        except StopIteration:
            self.destroy()
            return True
        except Exception as e:
            self.exception = e
            self.destroy()
            return True

    def _repr_args(self):
        args = IOHandler._repr_args(self)
        if not self.is_running:
            args.append("stopped")
        if self.recursion:
            args.append("recursion")
        return args


class LineOutputHandler(OutputHandler):
    def handle_event_non_recursive(self, event):
        try:
            val = next(self.iter)
            if val is not None:
                os.write(self.fd, val + b"\n")
            log.log("WRITE %s, %s" % (self.fd, repr(val)), "io")
        except StopIteration:
            log.log("STOP ITERATION %s" % self.fd, "ioevent")
            self.destroy()
            return True
        except Exception as e:
            self.exception = e
            self.destroy()
            return True

class InputHandler(IOHandler):
    events = select.POLLIN | select.POLLHUP | select.POLLERR
    
    def __init__(self, fd, borrowed = False, usage = None):
        self.buffer = None
        self.eof = False
        IOHandler.__init__(self, fd, borrowed, usage)

    def handle_event(self, event):
        if self.buffer is None:
            self.buffer = os.read(self.fd, 1024)
            if not self.buffer:
                self.eof = True
                self.destroy()
        return True

    def __iter__(self):
        return self

    def __next__(self):
        while self.buffer is None:
            if self.eof:
                raise StopIteration
            try:
                get_io_manager().handle_io()
            except RecursiveEvent:
                return None
        try:
            return self.buffer
        finally:
            self.buffer = None

    def _repr_args(self):
        args = IOHandler._repr_args(self)
        if self.eof:
            args.append("EOF")
        if self.buffer:
            args.append("data")
        return args

class LineInputHandler(InputHandler):
    def __init__(self, fd, borrowed = False, usage = None):
        InputHandler.__init__(self, fd, borrowed, usage)
        self.buffer = b""

    def handle_event(self, event):
        if b'\n' not in self.buffer:
            read_data = os.read(self.fd, 1024)
            self.buffer += read_data
            if not read_data:
                self.eof = True
                self.destroy()
        return True

    def __next__(self):
        while not self.eof and b'\n' not in self.buffer:
            try:
                get_io_manager().handle_io()
            except RecursiveEvent:
                return None
        if not self.buffer:
            assert self.eof
            raise StopIteration
        if b'\n' not in self.buffer:
            self.buffer += b'\n' # No newline at end of file...
        ret, self.buffer = self.buffer.split(b"\n", 1)
        return ret.decode("utf-8")

ALL_SIGNALS = set(getattr(signal, name) for name in dir(signal) if name.startswith("SIG") and '_' not in name)

class SignalManager(IOHandler):
    events = select.POLLIN
    
    def __init__(self, mask = [signal.SIGCHLD]):
        self.mask = mask
        self.signal_handlers = {}
        IOHandler.__init__(self, signalfd.signalfd(-1, mask, signalfd.SFD_CLOEXEC | signalfd.SFD_NONBLOCK), usage="SignalManager")
        signalfd.sigprocmask(signalfd.SIG_BLOCK, mask)

    def filter_to_key(self, flt):
        key = sorted(flt.items(), key=lambda item: item[0])
        return tuple(key)

    def register(self, signal_handler):
        flt = signal_handler.filter.items()

        key = self.filter_to_key(signal_handler.filter)
        self.signal_handlers[key] = signal_handler
        log.log("REGISTER %s, %s" % (key, signal_handler), "signalreg")

    def deregister(self, signal_handler):
        key = self.filter_to_key(signal_handler.filter)
        del self.signal_handlers[key]
        log.log("DEREGISTER %s, %s" % (key, signal_handler), "signalreg")

    def match_signal(self, siginfo, flt):
        for key, value in flt.items():
            if siginfo[key] != value:
                return False
        return True

    def handle_event(self, event):
        res = False
        while True:
            try:
                siginfo = signalfd.read_siginfo(self.fd)
            except (OSError, IOError) as e:
                if e.errno == errno.EAGAIN:
                    break
                raise

            siginfo = {name: getattr(siginfo, name) for name in dir(siginfo)}

            # Handle multiple simultaneously delivered SIGCHLD which
            # gets squashed into just one delivered signal event by
            # Linux...
            siginfos = [siginfo]
            if siginfo["ssi_signo"] == signal.SIGCHLD:
                siginfos = get_sigchlds()

            for siginfo in siginfos:
                if log.debug["signal"]:
                    log.log("Signal", "signal")
                    for key, val in siginfo_to_names(siginfo).items():
                        log.log("    %s: %s" % (key, val), "signal")

                for key, signal_handler in list(self.signal_handlers.items()):
                    if self.match_signal(siginfo, signal_handler.filter):
                        sighandlerres = signal_handler.handle_event(siginfo)
                        res = res or sighandlerres
        return res

    def _repr_args(self):
        args = IOHandler._repr_args(self)
        args.append(repr(self.mask))
        args.append(repr(self.signal_handlers))
        return args

CLD_EXITED = 1    # Child has exited.
CLD_KILLED = 2    # Child was killed.
CLD_DUMPED = 3    # Child terminated abnormally.
CLD_TRAPPED = 4   # Traced child has trapped.
CLD_STOPPED = 5   # Child has stopped.
CLD_CONTINUED = 6 # Stopped child has continued.

signals_by_value = {
    value: name
    for name, value in ((name, getattr(signal, name)) for name in dir(signal))
    if isinstance(value, int)                    
}

def siginfo_to_names(siginfo):
    siginfo = dict(siginfo)
    for key in siginfo:
        val = siginfo[key]
        if key == "ssi_code":
            val = {
                1: "CLD_EXITED",
                2: "CLD_KILLED",
                3: "CLD_DUMPED",
                4: "CLD_TRAPPED",
                5: "CLD_STOPPED",
                6: "CLD_CONTINUED"}.get(val, val)
        if (   key == "ssi_signo"
            or (key == "ssi_status"
                and siginfo["ssi_code"] != CLD_EXITED)):
            val = signals_by_value.get(val, val)
        siginfo[key] = val
    return siginfo

def get_sigchlds():
    try:
        while True:
            (pid, status) = os.waitpid(-1, os.WUNTRACED | os.WCONTINUED | os.WNOHANG)
            if pid == 0:
                raise StopIteration()

            res = {
                "ssi_signo": 0,   # Signal number
                "ssi_errno": 0,   # Error number (unused)
                "ssi_code": 0,    # Signal code
                "ssi_pid": 0,     # PID of sender
                "ssi_uid": 0,     # Real UID of sender
                "ssi_fd": 0,      # File descriptor (SIGIO)
                "ssi_tid": 0,     # Kernel timer ID (POSIX timers)
                "ssi_band": 0,    # Band event (SIGIO)
                "ssi_overrun": 0, # POSIX timer overrun count
                "ssi_trapno": 0,  # Trap number that caused signal
                "ssi_status": 0,  # Exit status or signal (SIGCHLD)
                "ssi_int": 0,     # Integer sent by sigqueue(3)
                "ssi_ptr": 0,     # Pointer sent by sigqueue(3)
                "ssi_utime": 0,   # User CPU time consumed (SIGCHLD)
                "ssi_stime": 0,   # System CPU time consumed (SIGCHLD)
                "ssi_addr": 0,    # Address that generated signal (for hardware-generated signals)
            }

            res["ssi_signo"] = signal.SIGCHLD
            res["ssi_pid"] = pid

            if os.WIFEXITED(status):
                res["ssi_code"] = CLD_EXITED
                res["ssi_status"] = os.WEXITSTATUS(status)
            elif os.WCOREDUMP(status):
                res["ssi_code"] = CLD_DUMPED
                res["ssi_status"] = os.WTERMSIG(status)
            elif os.WIFCONTINUED(status):
                res["ssi_code"] = CLD_CONTINUED
            elif os.WIFSTOPPED(status):
                res["ssi_code"] = CLD_STOPPED
                res["ssi_status"] = os.WSTOPSIG(status)
            elif os.WIFSIGNALED(status):
                res["ssi_code"] = CLD_KILLED
                res["ssi_status"] = os.WTERMSIG(status)

            yield res
    except OSError:
        raise StopIteration()

class SignalHandler(object):
    def __init__(self, filter):
        self.filter = filter
        get_signal_manager().register(self)
    def handle_event(self, event):
        pass
    def destroy(self):
        get_signal_manager().deregister(self)
        log.log("CLOSE DESTROY %s, %s" % (self.filter, self), "signalreg")
    
class SignalIteratorHandler(SignalHandler):
    def __init__(self, filter):
        SignalHandler.__init__(self, filter)
        self.buffer = []
    def handle_event(self, event):
        self.buffer[0:0] = [event]

    def __iter__(self):
        return self

    def __next__(self):
        while not self.buffer:
            get_io_manager().handle_io()
        return self.buffer.pop()

class ProcessSignalHandler(SignalHandler):
    def __init__(self, pid):
        self.last_event = None
        self.is_running = True
        self.pid = pid
        SignalHandler.__init__(self, {"ssi_pid": pid})

    def handle_event(self, event):
        log.log("Process event %s" % self.pid, "signal")
        self.last_event = event
        if event["ssi_code"] == CLD_EXITED:
            log.log("EXIT %s" % self.pid, "signal")
            self.is_running = False
            self.destroy()
            return True

# Generate an io-manager, hopefully for the main thread
# If we don't have one in the main thread, signal handling won't work.
get_io_manager()



import pkg_resources

from .iterio import *
from .redir import *
from .pipeline import *
from .environ import *
from .shell import *
from .module import *
from .utils import *
from .version import *
import builtins as __pieshell_builtins

name = "PieShell %s <)" % (version,)

with pkg_resources.resource_stream("pieshell", "README.md") as f:
    __doc__ = "%s\n\n%s" % (name, f.read().decode("utf-8"))

banner = """%s
Python %s
Type help(pieshell) for more information.""" % (name, sys.version.replace("\n", " "),)



import os
import fcntl
import types
import sys
import tempfile
import uuid
import code
import traceback
import threading
import signal
import signalfd
import operator
import re
import builtins        
import functools

from .. import copy
from .. import iterio
from .. import redir
from .. import log
from . import running
from . import base
from . import command

class BaseCommand(base.Pipeline):
    """This class is an abstract base class for all commands. It can
    however be instantiated, and will then return an instance of
    either Command or one of the Builtin classes.

    Command arguments are sent in as a list of strings and dictionaries.
    Elements of dictionary arguments are converted into --key=value
    pairs. Note that this short hand syntax might not work for all
    programs, as some expect "--key value", or even "-key=value" (e.g.
    find).

    You can work around this limitation by registering Command
    subclasses that specialcases this handling with the
    BuiltinRegistry.
    """
    def __new__(cls, env, arg = None):
        from . import builtin
        if cls is BaseCommand:
            if arg:
                builtin_cls = builtin.BuiltinRegistry.get_by_name(arg[0])
                if builtin_cls:
                    cls = builtin_cls
                else:
                    cls = Command
        return base.Pipeline.__new__(cls, env, arg)

    def __init__(self, env, arg = None):
        base.Pipeline.__init__(self, env)
        self._arg = arg and list(arg) or []
        self._running_process = None

    def __deepcopy__(self, memo = {}):
        return type(self)(self._env, copy.deepcopy(self._arg))

    def __call__(self, *arg, **kw):
        """Appends a set of arguments to the argument list

            env.mycommand("input_filename", verbose='3', destination="output")

        is equivalent to

            Command(env, ["mycommand", "input_filename", "--verbose=3", "--destination=output"])
        """
        arg = list(arg)
        if kw:
            arg += [kw]
        return type(self)(self._env, self._arg + arg)

    def __getattr__(self, name):
        """Append a name to the argument list, such that e.g.

            env.git.status("--help")

        is equivalent to

            env.git("status", "--help")
        """
        return type(self)(self._env, self._arg + [name])

    def _repr(self):
        args = self._arg or []

        for prefix_idx in range(0, len(args) + 1):
            if prefix_idx == len(args):
                break
            if not isinstance(args[prefix_idx], (bytes, str)) or not re.match(r"^[a-zA-Z]*$", args[prefix_idx]):
                break

        if prefix_idx:
            prefix = '.'.join(args[:prefix_idx])
        else:
            prefix = "_"
        args = args[prefix_idx:]

        args = [repr(arg) for arg in args]

        running_process = ''
        if self._running_process:
            running_process = ' as ' + repr(self._running_process)

        return u"%s(%s)%s" % (prefix, ', '.join(args), running_process)

    def _arg_list(self, redirects = None, sess = None, indentation = ""):
        orig_redirects = redir.Redirects(redirects)
        orig_redirects.borrow()
        def handle_arg_pipes(item):
            if isinstance(item, str):
                return item
            elif redirects is not None:
                return self._handle_arg_pipes(item, orig_redirects, redirects, sess, indentation)
            else:
                return "/dev/fd/X"
        args = []
        if self._arg:
            for arg in self._arg:
                if isinstance(arg, dict):
                    for name, value in arg.items():
                        for match in self._env._expand_argument(handle_arg_pipes(value)):
                            args.append("--%s=%s" % (name, match))
                else:
                    for match in self._env._expand_argument(handle_arg_pipes(arg)):
                        args.append(match)
        return args

    def _arg_list_sh(self, *arg, **kw):
        def quote_arg(arg):
            arg = str(arg)
            if " " in arg:
                arg = repr(arg)
            return arg
        return ' '.join(quote_arg(arg) for arg in self._arg_list(*arg, **kw))

    def _run(self, redirects, sess, indentation = ""):
        raise NotImplemented


class Command(command.BaseCommand):
    """Runs an external program with the specified arguments.
    Arguments are sent in as a list of strings and dictionaries.
    Elements of dictionary arguments are converted into --key=value
    pairs. Note that this short hand syntax might not work for all
    programs, as some expect "--key value", or even "-key=value" (e.g.
    find). """

    def _child(self, redirects, args):
        redirects.perform()
        os.chdir(self._env._cwd)
        os.execvpe(args[0], args, self._env._exports)
        os._exit(-1)

    def _handle_arg_pipes(self, thing, orig_redirects, redirects, sess, indentation):
        from . import function
        if isinstance(thing, str):
            return thing
        elif isinstance(thing, base.Pipeline):
            direction = "stdout"
        elif isinstance(thing, (types.FunctionType, types.MethodType)):
            thing = function.Function(self._env, thing)
            direction = "stdin"
        elif hasattr(thing, "__iter__") or hasattr(thing, "__next__"):
            thing = function.Function(self._env, thing)
            direction = "stdout"
        else:
            # Not a named pipe item, just a string
            return thing
      
        arg_pipe = thing._run(
            redir.Redirects(
                orig_redirects,
                redir.Redirect(direction, redir.PIPE)),
            sess,
            indentation + "  ")

        fd = redirects.find_free_fd()
        redirects.redirect(
            fd,
            getattr(thing._redirects, direction).pipe,
            {"stdin": os.O_WRONLY, "stdout": os.O_RDONLY}[direction])

        self._running_processes.extend(arg_pipe)

        return "/dev/fd/%s" % fd

    def _run(self, redirects, sess, indentation = ""):
        base.Pipeline._run(self, redirects, sess, indentation)

        self._running_processes = []

        redirects = redirects.make_pipes()
        log.log(indentation + "Running %s with %s" % (repr(self), repr(redirects)), "cmd")

        args = self._arg_list(redirects, sess, indentation)
        log.log(indentation + "222Running %s with %s" % (repr(self), repr(redirects)), "cmd")

        pid = os.fork()
        if pid == 0:
            self._child(redirects, args)
            # If we ever get to here, all is lost...
            sys._exit(-1)

        log.log(indentation + "  %s: Command line %s with %s" % (pid, ' '.join(repr(arg) for arg in args), repr(redirects)), "cmd")

        self._running_process = running.RunningProcess(self, pid)
        self._running_processes.append(self._running_process)

        redirects.close_source_fds()

        self._pid = pid
        self._redirects = self._running_process.redirects = redirects

        return self._running_processes

    def _complete(self):
        cmd = self._arg_list_sh() + " "
        return (item.strip() for item in self._env.get_completions(cmd))

    def __dir__(self):
        try:
            return list(self._complete())
        except Exception as e:
            traceback.print_exc()
            return ["<%s>" % e]

    @property
    def __doc__(self):
        return b"\n".join(self("--help")).decode("utf-8")



import os
import sys
import traceback
import glob
import code

from . import pipeline
from . import redir
import pieshell


class R(object):
    """Wraps a string and protects it from path expansions"""
    def __init__(self, str):
        self.str = str
    def __getattr__(self, name):
        return getattr(self.str, name)
    def __iter__(self):
        return iter(self.str)
    def __repr__(self):
        return "R(%s)" % (repr(self.str),)

class Environment(object):
    """An environment within which a command or pipeline can run. The
    environment consists of a current working directory and a set of
    environment variables and other configuration.

    Commands within the environment can be convienently created using
    the

        env.COMMAND_NAME

    as a short hand for

        Command(env, "COMMAND_NAME")
    """

    def __init__(self, cwd = None, exports = None, interactive = False, redirects = None):
        """Creates a new environment from scratch. Takes the same
        arguments as __call__."""
        self._cwd = os.getcwd()
        if cwd is not None:
            self._cd(cwd)
        self._exports = exports
        self._interactive = interactive
        self._bashfunctions = {}
        self._scope = None
        if redirects is None:
            redirects = redir.Redirects()
            redirects.redirect(0, 0, borrowed=True)
            redirects.redirect(1, 1, borrowed=True)
            redirects.redirect(2, redir.STRING)
        self._redirects = redirects
        self._clear_dir_cache()
    @property
    def _exports(self):
        return self._exports_value or os.environ
    @_exports.setter
    def _exports(self, exports):
        self._exports_value = exports
    @_exports.deleter
    def _exports(self):
        self._exports_value = None
    def _expand_path(self, pth):
        if not pth.startswith("/") and not pth.startswith("~"):
            pth = os.path.join(self._cwd, pth)
        pth = os.path.expanduser(pth)
        pth = os.path.abspath(pth)
        return pth
    def _expand_argument(self, arg):
        """Performs argument glob expansion on an argument string.
        Returns a list of strings.
        """
        if isinstance(arg, R): return [arg.str]
        scope = self._scope or self._exports
        arg = arg % scope
        arg = os.path.expanduser(arg)
        res = glob.glob(self._expand_path(arg))
        if not res: return [arg]
        if self._cwd != "/":
            for idx in range(0, len(res)):
                if res[idx].startswith(self._cwd + "/"):
                    res[idx] = "./" + res[idx][len(self._cwd + "/"):]
        return res
    def _cd(self, cwd):
        cwd = self._expand_path(cwd)
        if not os.path.exists(cwd):
            raise IOError("Path does not exist: %s" % cwd)
        self._cwd = cwd
        if self._interactive:
            os.chdir(cwd)
        return self
    def __call__(self, cwd = None, exports = None, interactive = None, redirects = None):
        """Creates a new environment based on the current ones. All
        configuration is copied, unless specifically overridden.

        cwd: Path to current working directory
        exports: Dictionary of environment variables
        interactive: Boolean. If true:
            * Changing the current working directory of this
              environment changes the real working directory of the
              current python process using os.chdir().
            * repr() of a pipeline will run the pipeline with
              stdin/stdout/stderr connected to the current terminal,
              and wait until the pipeline terminates.
        """
        if exports is None:
            exports = self._exports
        if interactive is None:
            interactive = self._interactive
        if redirects is None:
            redirects = self._redirects
        res = type(self)(cwd = self._cwd, exports = exports, interactive = interactive, redirects = redirects)
        if cwd is not None:
            res.cd(cwd)
        return res
    def __getitem__(self, name):
        """env[path] is equivalent to env(path)"""
        return self(name)
    def __getattr__(self, name):
        """Creates a pipeline of one command in the current
        environment."""
        if name == "_":
            return pipeline.command.BaseCommand(self)
        else:
            return pipeline.command.BaseCommand(self, [name])
    def _coerce(self, thing, direction):
        if thing is None:
            thing = "/dev/null"
        if isinstance(thing, (str, unicode)):
            thing = redir.Redirect(direction, thing)
        if isinstance(thing, redir.Redirect):
            thing = redir.Redirects(thing, defaults=False)
        if not isinstance(thing, redir.Redirect):
            raise ValueError(type(thing))
        return thing
    def __ror__(self, other):
        """Sets default redirects."""
        self._redirects.register(self._redirects._coerce(other, 'stdin'))
    def __or__(self, other):
        """Sets default redirects."""
        self._redirects.register(self._redirects._coerce(other, 'stdout'))
    def __repr__(self):
        """Prints the current prompt"""
        if self._interactive:
            return "%s:%s >>> " % (str(id(self))[:3], self._cwd)
        else:
            return "[%s:%s]" % (str(id(self))[:3], self._cwd)
    def _clear_dir_cache(self):
        self._dir_cache = None
    def __dir__(self):
        if self._dir_cache is None:
            e = self._exports
            self._dir_cache = []
            paths = e["PATH"].split(":")
            for pth in paths:
                if not pth.startswith("/"):
                    pth = os.path.join(self._cwd, pth)
                self._dir_cache.extend(os.listdir(os.path.abspath(pth)))
            self._dir_cache.extend(pipeline.BuiltinRegistry.builtins.keys())
            self._dir_cache.sort()
        return self._dir_cache

env = Environment()

class EnvScope(dict):
    """EnvScope can be used instead of a globals() dictionary to allow
    global lookup of command names, without the env.COMMAND
    prefixing."""
    def __setitem__(self, name, value):
        # Hack for ptpython
        if name == "_":
            name = "last_statement"
        if name == "env":
            value._scope = self
        env = dict.__getitem__(self, 'env')
        if name in env._exports:
            env._exports[name] = value
        else:
            dict.__setitem__(self, name, value)
    def __getitem__(self, name):
        try:
            return dict.__getitem__(self, name)
        except KeyError:
            pass
        env = dict.__getitem__(self, 'env')
        if name != "_":
            if name == "exports":
                return env._exports
            if name in env._exports:
                return env._exports[name]
            if name in __builtins__:
                raise KeyError(name)
        return getattr(env, name)

    def keys(self):
        return dict.keys(self) + dir(dict.__getitem__(self, 'env'))

    def __bytes__(self):
        return str(self).encode("utf-8")
    
    def __str__(self):
        try:
            return str(dict.__getitem__(self, 'env'))
        except Exception as e:
            traceback.print_exc()
            return '<%s>' % e

    def execute_file(self, filename):
        with open(filename) as f:
            content = f.read()
        code.InteractiveConsole(locals=self).runsource(content, filename, "exec")

    def execute_expr(self, expr):
        exec(expr, self)

    def execute_startup(self):
        env = self["env"]
        self.execute_expr("from pieshell import *")
        self["env"] = env
        self.execute_expr("import readline")
        conf = os.path.expanduser('~/.config/pieshell')
        if os.path.exists(conf):
            self.execute_file(conf)

    def __enter__(self):
        self.ps1 = getattr(sys, "ps1", None)
        sys.ps1 = self

    def __exit__(self, *args, **kw):
        sys.ps1 = self.ps1

envScope = EnvScope(env = env(interactive=True))



import os
import fcntl
import types
import sys
import tempfile
import uuid
import code
import traceback
import threading
import signal
import signalfd
import operator
import re
import builtins        
import functools

from .. import copy
from .. import redir
from .. import log
from . import running

repr_state = threading.local()
standard_repr = builtins.repr
def pipeline_repr(obj):
    """Returns a string representation of an object, including pieshell
    pipelines."""

    if not hasattr(repr_state, 'in_repr'):
        repr_state.in_repr = 0
    repr_state.in_repr += 1
    try:
        return standard_repr(obj)
    finally:
        repr_state.in_repr -= 1
builtins.repr = pipeline_repr

# help() doesn't let you override help on objects, only on classes, so
# make everything a class...
class DescribableObject(type):
    def __new__(cls, *arg, **kw):
        return type.__new__(cls, "", (type,), {})
    def __init__(self, *arg, **kw):
        pass

class Pipeline(DescribableObject):
    """Abstract base class for all pipelines"""

    _print_state = threading.local()
    def __init__(self, env):
        self._env = env
        self._started = False
    def __deepcopy__(self, memo = {}):
        return type(self)(self._env)
    def _coerce(self, thing, direction):
        from . import function
        if thing is None:
            thing = "/dev/null"
        if isinstance(thing, (str, bytes)):
            thing = redir.Redirect(direction, thing)
        if isinstance(thing, redir.Redirect):
            thing = redir.Redirects(thing, defaults=False)
        if not isinstance(thing, Pipeline) and (isinstance(thing, types.FunctionType) or hasattr(thing, "__iter__") or hasattr(thing, "__next__")):
            thing = function.Function(self._env, thing)
        if not isinstance(thing, (Pipeline, redir.Redirects)):
            raise ValueError(type(thing))
        return thing
    def __ror__(self, other):
        """Pipes the standard out of a pipeline into the standrad in
        of this pipeline."""
        from . import redirect
        from . import pipe
        other = self._coerce(other, 'stdin')
        if isinstance(other, redir.Redirects):
            return redirect.CmdRedirect(self._env, self, other)
        else:
            return pipe.Pipe(self._env, other, self)
    def __or__(self, other):
        """Pipes the standard out of the pipeline into the standrad in
        of another pipeline."""
        from . import redirect
        from . import pipe
        other = self._coerce(other, 'stdout')
        if isinstance(other, redir.Redirects):
            return redirect.CmdRedirect(self._env, self, other)
        else:
            return pipe.Pipe(self._env, self, other)
    def __gt__(self, file):
        """Redirects the standard out of the pipeline to a file."""
        return self | file
    def __lt__(self, file):
        """Redirects the standard in of the pipeline from a file."""
        return file | self
    def __add__(self, other):
        return Group(self._env, self, other)

    def _run(self, redirects, sess, indentation = ""):
        self._started = True

    def run(self, redirects = []):
        """Runs the pipelines with the specified redirects and returns
        a RunningPipeline instance."""
        if not isinstance(redirects, redir.Redirects):
            redirects = redir.Redirects(self._env._redirects, *redirects)
        with copy.copy_session() as sess:
            self = copy.deepcopy(self)
            processes = self._run(redirects, sess)
        pipeline = running.RunningPipeline(processes, self)
        self._env.last_pipeline = pipeline
        return pipeline

    def run_interactive(self):
        pipeline = self.run()
        try:
            pipeline.wait()
        except KeyboardInterrupt as e:
            raise PipelineInterrupted(pipeline)
        return pipeline

    def __iter__(self):
        """Runs the pipeline and iterates over its standrad output lines."""
        return iter(self.run([redir.Redirect("stdout", redir.PIPE)]))
    def __str__(self):
        # FIXME: Should use locale, but python's locale module is broken and ignores LC_* by default
        return bytes(self).decode("utf-8")
    def __bytes__(self):
        """Runs the pipeline and returns its standrad out output as a string"""
        return "\n".join(iter(self.run([redir.Redirect("stdout", redir.PIPE)])))
    def __invert__(self):
        """Start a pipeline in the background"""
        return self.run()
    def __repr__(self):
        """Runs the command if the environment has interactive=True,
        sending the output to standard out. If the environment is
        non-interactive, returns a string representation of the
        pipeline without running it."""

        if not self._started and self._env._interactive and getattr(repr_state, "in_repr", 0) < 1:
            try:
                self.run_interactive()
            except KeyboardInterrupt as e:
                log.log("Canceled:\n%s" % (e,), "error")
            except Exception as e:
                log.log("Error:\n%s" % (e,), "error")
                sys.last_traceback = sys.exc_info()[2]
                import pdb
                pdb.pm()
            return ''
        else:
            current_env = getattr(Pipeline._print_state, 'env', None)
            Pipeline._print_state.env = self._env
            try:
                envstr = ''
                if current_env is not self._env:
                    envstr = repr(self._env)
                return "%s%s" % (envstr, self._repr())
            finally:
                Pipeline._print_state.env = current_env

    def __dir__(self):
        return []

    @property
    def __bases__(self):
        return []

    @property
    def __name__(self):
        current_env = getattr(Pipeline._print_state, 'env', None)
        Pipeline._print_state.env = self._env
        try:
            return repr(self)
        finally:
            Pipeline._print_state.env = current_env

    @property
    def __doc__(self):
        return ""



from . import command

class BuiltinRegistry(object):
    builtins = {}

    @classmethod
    def register(cls, builtin_cls):
        cls.builtins[builtin_cls.name] = builtin_cls

    @classmethod
    def get_by_name(cls, name):
        if name not in cls.builtins:
            return None
        return cls.builtins[name]

class Builtin(command.BaseCommand):
    def _run(self, redirects, sess, indentation = ""):
        raise NotImplemented



import os
import sys

debug = {
    "all": False,
    "error": True,
    "fd": False,
    "cmd": False,
    "ioreg": False,
    "ioevent": False,
    "io": False,
    "signalreg": False,
    "signal": False,
    "test": False
    }

outfd = None
if hasattr(sys.stdout, "fileno"):
    outfd = sys.stdout.fileno()
elif hasattr(sys.stderr, "fileno"):
    outfd = sys.stderr.fileno()
else:
    outfd = 1
logfd = 1023
os.dup2(outfd, logfd)
def log(msg, category="misc"):
    if not debug.get(category, False) and not debug.get("all", False): return
    os.write(logfd, ("%s: %s\n" % (os.getpid(), msg)).encode("utf-8"))



import sys
import os
import code

from . import environ
from . import log
from . import version

# Example usage
# for line in env.find(".", name='foo*', type='f') | env.grep("bar.*"):
#    print line

def main():

    def test():
        try:
            e = env
            print("===={test one}====")
            for x in e.ls | e.grep(".py$") | e.sed("s+shell+nanan+g"):
                print(x)

            print("===={test two}====")
            def somefn():
                yield "foo bar fien\n"
                yield "foo naja hehe\n"
                yield "bar naja fie\n"

            for x in somefn() | e.grep("foo"):
                print(x)


            print("===={test three}====")
            data = [
                "foo bar fien\n",
                "foo naja hehe\n",
                "bar naja fie\n"
                ]

            print(list(data | e.grep("foo")))

            # print("===={test four}====")

            # for x in ((e.echo("hejjo") | e.sed("s+o+FLUFF+g"))
            #            + e.echo("hopp")
            #          ) | e.sed("s+h+nan+g"):
            #     print(x)

            print("===={test five}====")

            print(list(env.cat(iter(["foo", "bar", "fie"])) | env.cat()))

        except:
            import sys, pdb
            sys.last_traceback = sys.exc_info()[2]
            pdb.pm()

    args = []
    kws = {}
    for arg in sys.argv[1:]:
        if arg.startswith("--"):
            arg = arg[2:]
            if "=" in arg:
                name, value = arg.split("=", 1)
            else:
                name = arg
                value = True
            kws[name] = value
        else:
            args.append(arg)

    if kws.get("help", False):
        print("""Usages:

pieshell
pieshell FILE.pysh
pieshell --cmd='any valid pieshell command or python statement'
pieshell --test
pieshell --help
pieshell --version
pieshell --ptpython
    Fancy editing environment based on ptpython (pip install ptpython)
pieshell --log=NAME,NAME,NAME


""")
    elif kws.get("version", False):
        print(version.version)
    elif kws.get("test", False):
        test()
    else:
        if 'log' in kws:
            for name in kws['log'].split(','):
                log.debug[name] = True

        with environ.envScope:
            environ.envScope["args"] = args
            environ.envScope["kws"] = kws
            environ.envScope.execute_startup()

            if "cmd" in kws:
                environ.envScope.execute_expr(kws["cmd"])
            elif args:
                for arg in args:
                    environ.envScope.execute_file(arg)
            else:
                if kws.get("ptpython", False):
                    import pieshell.monkeypatches.patch_jedi
                    import ptpython.repl

                    import pygments.token
                    import ptpython.prompt_style
                    def in_tokens(self, cli):
                        return [(pygments.token.Token.Prompt, str(environ.envScope))]
                    ptpython.prompt_style.ClassicPrompt.in_tokens = in_tokens

                    ptpython.repl.embed(locals=environ.envScope, vi_mode=False)
                else:
                    import pieshell
                    code.InteractiveConsole(locals=environ.envScope).interact(banner=pieshell.banner, exitmsg="...om nom nom")

if __name__ == '__main__':
    main()



import sys
import os
import types
from . import environ

class Module(types.ModuleType):
    def __init__(self, scope, name, doc = None):
        types.ModuleType.__setattr__(self, "_scope", scope)
        super(Module, self).__init__(name, doc)
    def __getattr__(self, name):
        return self._scope[name]
    def __setattr__(self, name, value):
        self._scope[name] = value
    def __dir__(self):
        return dict.keys(self._scope)

class Loader(object):
    def __init__(self, path):
        self.path = path
    def load_module(self, fullname):
        if fullname not in sys.modules:
            env = environ.envScope["env"]()
            scope = environ.EnvScope(env=env)
            sys.modules[fullname] = Module(scope, fullname.split(".")[-1])
        sys.modules[fullname]._scope.execute_file(self.path)
        return sys.modules[fullname]

class Finder(object):
    def find_module(self, fullname, path=None):
        filename = fullname.split(".")[-1] + ".pysh"
        paths = sys.path
        if path is not None:
            paths = path
        if not isinstance(paths, (list, tuple)):
            paths = [paths]
        found = None
        for path in paths:
            if path == "": path = "."
            filepath = os.path.join(path, filename)
            if os.path.exists(filepath):
                found = filepath
                break
        if found is None:
            return None
        return Loader(found)

sys.meta_path.append(Finder())



import os
import fcntl
import types
import sys
import tempfile
import uuid
import code
import traceback
import threading
import signal
import signalfd
import operator
import re
import builtins        
import functools

from .. import copy
from .. import iterio
from .. import redir
from .. import log
from . import running
from . import base
from . import command
from . import pipe

class Function(base.Pipeline):
    """Encapsulates a function or iterator so that it can be used
    inside a pipeline. An iterator can only have its output piped into
    something. A function can have its output piped into something by
    yeilding values, and can take input in the form of an iterator as
    a sole argument."""

    def __init__(self, env, function, *arg, **kw):
        base.Pipeline.__init__(self, env)
        self.function = function
        self._arg = arg
        self._kw = kw
    def __deepcopy__(self, memo = {}):
        return type(self)(self._env, self.__dict__["function"], *copy.deepcopy(self._arg), **copy.deepcopy(self._kw))
    def _function_name(self):
        thing = self.__dict__["function"] # Don't wrap functions as instance methods
        if isinstance(thing, (types.FunctionType, types.MethodType)):
            mod = thing.__module__ or ''
            if mod:
                mod = mod + '.'
            
            if isinstance(thing, types.MethodType):
                func = "%s.%s" % (type(thing.__self__).__name__, thing.__name__)
            else:
                func = thing.__name__

            return u"%s%s" % (mod, func)
        else:
            return repr(thing)
    def _repr(self):
        thing = self.__dict__["function"] # Don't wrap functions as instance methods
        if isinstance(thing, (types.FunctionType, types.MethodType)):
            args = []
            if self._arg:
                args += [repr(arg) for arg in self._arg]
            if self._kw:
                args += ["%s=%s" % (key, repr(value)) for (key, value) in self._kw.items()]
            return u"%s(%s)" % (self._function_name(), ','.join(args))
        else:
            return repr(thing)

    def _run(self, redirects, sess, indentation = ""):
        base.Pipeline._run(self, redirects, sess, indentation)
        
        output_iterpipe = None
        if (    isinstance(redirects.stdout.source, type)
            and issubclass(redirects.stdout.source, pipe.ITERPIPE)):
            output_iterpipe = redirects.stdout.source

        if output_iterpipe:
            redirects.redirect("stdout")

        redirects = redirects.make_pipes()
        log.log(indentation + "Running %s with %s" % (repr(self), repr(redirects)), "cmd")

        def convert(x):
            if x is None:
                return x
            elif isinstance(x, bytes):
                return x
            elif isinstance(x, str):
                return x.encode("utf-8")
            else:
                return str(x).encode("utf-8")

        thing = self.__dict__["function"] # Don't wrap functions as instance methods
        if isinstance(thing, (types.FunctionType, types.MethodType)):
            if isinstance(redirects.stdin.source, pipe.ITERPIPE):
                thing = thing(redirects.stdin.source.iter, *self._arg, **self._kw)
            else:
                thing = thing(
                    iterio.LineInputHandler(redirects.stdin.open(False), usage=self),
                    *self._arg, **self._kw)
        if hasattr(thing, "__iter__"):
            thing = iter(thing)

        if output_iterpipe:
            self._running_processes = []
            self._redirects = redirects
            self._redirects.redirect(
                "stdout",
                source=output_iterpipe,
                pipe=output_iterpipe(iter=thing))
        else:
            self._running_process = running.RunningFunction(
                self,
                iterio.LineOutputHandler(
                    redirects.stdout.open(False),
                    (convert(x) for x in thing),
                    usage=self))
            self._running_processes = [self._running_process]
            self._redirects = self._running_process.redirects = redirects

        redirects.close_source_fds()

        return self._running_processes



import os
import fcntl
import types
import sys
import tempfile
import uuid
import code
import traceback
import threading
import signal
import signalfd
import operator
import re
import builtins        
import functools

from .. import copy
from .. import iterio
from .. import redir
from .. import log
from . import running
from . import base
from . import command
from . import function
from . import builtins
from . import pipe

class CmdRedirect(base.Pipeline):
    def __init__(self, env, pipeline, redirects):
        base.Pipeline.__init__(self, env)
        self.pipeline = pipeline
        self.cmd_redirects = redirects
    def __deepcopy__(self, memo = {}):
        return type(self)(self._env, copy.deepcopy(self.pipeline), copy.deepcopy(self.cmd_redirects))
    def _repr(self):
        return u"%s with %s" % (repr(self.pipeline), repr(self.cmd_redirects))
    def _run(self, redirects, sess, indentation = ""):
        base.Pipeline._run(self, redirects, sess, indentation)

        log.log(indentation + "Running [%s] with %s" % (repr(self), repr(redirects)), "cmd")

        res = self.pipeline._run(redirects.merge(self.cmd_redirects), indentation + "  ")
        self._redirects = self.pipeline._redirects
        return res



#! /usr/bin/python

from setuptools import setup, find_packages
import os.path

VERSION = "0.1.2"

with open(os.path.join(os.path.dirname(__file__), "README.md"), "r") as rf:
    with open(os.path.join(os.path.dirname(__file__), "pieshell", "README.md"), "w") as wf:
        wf.write(rf.read())

with open(os.path.join(os.path.dirname(__file__), "pieshell", "version.py"), "w") as wf:
    wf.write("version = '%s'\n" % (VERSION,)) 
    
setup(
    name = "pieshell",
    description = "Pieshell is a Python shell environment that combines the expressiveness of shell pipelines with the prower of python iterators. It can be used both as an interactive shell and as an ordinary python module replacing e.g. subprocess.Popen",
    keywords = "Python shell pipelines suprocess",
    install_requires = ["signalfd"],
    version = VERSION,
    author = "Egil Moeller",
    author_email = "egil.moller@piratpartiet.se",
    license = "GPL",
    url = "https://github.com/redhog/pieshell",
    packages = find_packages(),
    package_data={'pieshell': ['*.md']},
    entry_points={
        'console_scripts': [
            'pieshell = pieshell.shell:main',
        ],
    },
    scripts = ["contrib/get_completions"]
)



import os
import os.path
import shlex
import io

from . import builtin

class CdBuiltin(builtin.Builtin):
    """Change directory to the supplied path.
    """
    name = "cd"

    @property
    def _path(self):
        pth = "~"
        if self._arg[1:]:
            pth = os.path.join(*self._arg[1:])
        return pth

    def _run(self, redirects, sess, indentation = ""):
        self._env._cd(self._path)
        return []

    def __dir__(self):
        if self._arg[1:]:
            pth = self._env._expand_path(self._path)
        else:
            pth = "."
        try:
            return [name for name in os.listdir(pth)
                    if os.path.isdir(os.path.join(pth, name))]
        except:
            return []

builtin.BuiltinRegistry.register(CdBuiltin)

class ClearDirCacheBuiltin(builtin.Builtin):
    """Clear the tab completion cache
    """
    name = "clear_dir_cache"

    def _run(self, redirects, sess, indentation = ""):
        self.env._clear_dir_cache()
        return []

    def __dir__(self):
        if self._arg[1:]:
            pth = self.env._expand_path(self._path)
        else:
            pth = "."
        try:
            return [name for name in os.listdir(pth)
                    if os.path.isdir(os.path.join(pth, name))]
        except:
            return []

builtin.BuiltinRegistry.register(ClearDirCacheBuiltin)

def parse_declares(data):
    l = shlex.shlex(io.StringIO(data), posix=True)

    l.whitespace_split = True

    lastlineno = 1
    line = []
    lines = []
    while True:
        if l.lineno != lastlineno:
            lines.append((lastlineno, line))
            line = []
        lastlineno = l.lineno
        token = l.get_token()
        if token is None:
            break
        line.append(token)

    vars = {}
    funcstarts = []
    for (lineno, line) in lines:
        if line[0] == "declare":
            name = line[2]
            if "=" in name:
                name, value = name.split("=", 1)
            else:
                value = ""
            vars[name] = value
        elif len(line) == 3 and line[1] == '()' and line[2] == '{':
            funcstarts.append((lineno, line[0]))

    lines = data.split("\n")
    funcs = {}
    for i in range(len(funcstarts)):
        if i >= len(funcstarts) - 1:
            flines = lines[funcstarts[i][0]-1:]
        else:
            flines = lines[funcstarts[i][0]-1:funcstarts[i+1][0]-1]
        funcs[funcstarts[i][1]] = "\n".join(flines)

    return vars, funcs

class BashSource(builtin.Builtin):
    """Runs a bash script and imports all environment variables at the
    end of the script.
    """

    name = "bashsource"

    def _run(self, redirects, sess, indentation = ""):
        self._cmd = self._env.bash(
            "-l", "-i", "-c",
            "source '%s'; { declare -x; declare -f; } > $0" % (self._env._expand_argument(self._arg[1])[0],),
            self.parse_decls)
        res = self._cmd._run(redirects, sess, indentation)
        self._pid = self._cmd._pid
        self._redirects = self._cmd._redirects
        return res
    
    def parse_decls(self, stdin):
        # Parse and load environment variables from bash
        lines = []
        for line in stdin:
            if line is None:
                yield; continue
            lines.append(line)
        lines = "\n".join(lines)
        vars, funcs = parse_declares(lines)
        self._env._exports.update(vars)
        self._env._bashfunctions.update(funcs)
        yield

builtin.BuiltinRegistry.register(BashSource)



from .running import *
from .base import *
from .command import *
from .function import *
from .builtins import *
from .pipe import *
from .redirect import *

# class Group(base.Pipeline):
#     def __init__(self, env, first, second):
#         base.Pipeline.__init__(self, env)
#         self.first = first
#         self.second = second
#     def thread_main(self, stdin = None, stdout = None, stderr = None, *arg, **kw):
#         for item in [self.first, self.second]:
#             item.run(stdin=stdin, stdout=stdout, stderr=stderr, **kw).join()
#     def _repr(self):
#         return u"%s + %s" % (repr(self.first), repr(self.second))
